# frozen_string_literal: true

require "core/pipeline/callable"

require "is/inspectable"

require_relative "compiler"

module Core
  module Event
    # [public] Manages and calls an event pipeline.
    #
    class Callable
      include Is::Inspectable
      inspects without: [:@object, :@mutex, :@pipelines]

      def initialize(object)
        @object = object
        @mutex = Mutex.new
        @events = []
        @pipelines = {}
      end

      attr_reader :events
      attr_reader :pipelines

      # [public] Registers one or more events by name.
      #
      def register(*events)
        @mutex.synchronize do
          @events.concat(events.map(&:to_sym)).uniq!
        end
      end

      # [public] Defines a callback to be called during, before, and/or after the given events.
      #
      def callback(*during, before: [], after: [], method: nil, context: nil, &block)
        define_callback(:before, *before, method: method, context: context, &block)
        define_callback(:during, *during, method: method, context: context, &block)
        define_callback(:after, *after, method: method, context: context, &block)
      end

      private def define_callback(type, *defined_events, method: nil, context: nil, &block)
        defined_events.each do |defined_event|
          @mutex.synchronize do
            defined_event = defined_event.to_sym

            @pipelines[defined_event] ||= {
              before: Core::Pipeline::Callable.new(@object),
              during: Core::Pipeline::Callable.new(@object),
              after: Core::Pipeline::Callable.new(@object)
            }

            @pipelines[defined_event][type].action(method, context: context, &block)
          end
        end
      end

      # [public] Calls callbacks for the given event with the given arguments.
      #
      def performing(object, event, ...)
        compile.performing(object, event, ...)
      end

      # [public] Calls before callbacks for the given event with the given arguments.
      #
      def starting(object, event, ...)
        compile.starting(object, event, ...)
      end

      # [public] Calls after callbacks for the given event with the given arguments.
      #
      def finished(object, event, ...)
        compile.finished(object, event, ...)
      end

      private def compile
        instance_eval(Compiler.compile(self), __FILE__, __LINE__ - 1)

        self
      end

      private def call_callbacks(object, type, event, ...)
        validate_event(event)

        if (pipeline = @pipelines.dig(event.to_sym, type))
          pipeline.call(object, ...)
        end
      end

      private def validate_event(event)
        unless @events.include?(event)
          raise "unknown event `#{event}`"
        end
      end
    end
  end
end
